/*
 *  user_gpio.c -- Support users to use GPIO
 *
 *  Author              pxh
 *  Email               peixiuhui@163.com
 *  Create time         2015-08-14
 *  kernel version		3.10.53
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 */
#include <linux/module.h> 
#include <linux/init.h>
#include <linux/delay.h> 
#include <asm/irq.h> 
#include <linux/kernel.h> 
#include <linux/mm.h> 
#include <linux/fs.h> 
#include <linux/types.h> 
#include <linux/delay.h> 
#include <linux/slab.h> 
#include <linux/errno.h> 
#include <linux/ioctl.h> 
#include <linux/cdev.h> 
#include <asm/uaccess.h> 
#include <asm/atomic.h> 
#include <asm/unistd.h> 
#include <linux/platform_device.h>
#include <linux/io.h>
#include <linux/gpio.h>
#include <asm/gpio.h>
#include <linux/of.h>
#include <linux/of_gpio.h>
#include <linux/of_platform.h>

//#define GPIO_IOC_MAGIC					'G'
//#define IOCTL_GPIO_SETOUTPUT			_IOW(GPIO_IOC_MAGIC, 0, int)
//#define IOCTL_GPIO_SETINPUT				_IOW(GPIO_IOC_MAGIC, 1, int)
//#define IOCTL_GPIO_SETVALUE				_IOW(GPIO_IOC_MAGIC, 2, int)
//#define IOCTL_GPIO_GETVALUE				_IOR(GPIO_IOC_MAGIC, 3, int)
//#define GPIO_IOC_MAXNR					3

#ifndef IMX_GPIO_NR
#define IMX_GPIO_NR(bank, nr) ((bank-1)*32 + (nr))
#endif

//LED 开关定义
#define IOCTL_LED_ON	1
#define IOCTL_LED_OFF	0

struct s3c2416_gpio
{
	unsigned short pin;
	unsigned short value;
};


typedef struct {
	int pin;
	int data;
	int usepullup;
} user_gpio_arg;
 

struct user_gpio {
	unsigned int *gpios;
	int gpio_num;
	struct cdev gpio_cdev;
};

static int gpio_major = 0;
static int gpio_minor = 0;
static struct class *gpio_class;
static struct device *class_dev;

#define GPIO_PIN_SIZE 14

/* 用来指定LED所用的GPIO引脚 */
static const unsigned long gpio_table [GPIO_PIN_SIZE] =
{
	IMX_GPIO_NR(4, 19),		//00 DO0 高电平有效 MX6UL_PAD_CSI_VSYNC__GPIO4_IO19
	IMX_GPIO_NR(4, 20),		//01 DO1 高电平有效 MX6UL_PAD_CSI_HSYNC__GPIO4_IO20
	IMX_GPIO_NR(1, 18),		//02 DI0  MX6UL_PAD_UART1_CTS_B__GPIO1_IO18
	IMX_GPIO_NR(1, 23),		//03 DI1  MX6UL_PAD_UART2_RTS_B__GPIO1_IO23
	IMX_GPIO_NR(1, 22),		//04 DI2  MX6UL_PAD_UART2_CTS_B__GPIO1_IO22
	IMX_GPIO_NR(1, 27),		//08 DI3  MX6UL_PAD_UART3_RTS_B__GPIO1_IO27
	IMX_GPIO_NR(1, 26),		//09 DI4  MX6UL_PAD_UART3_CTS_B__GPIO1_IO26
	IMX_GPIO_NR(2, 14),		//05 DI5  MX6UL_PAD_ENET2_TX_CLK__GPIO2_IO14
	IMX_GPIO_NR(2, 8),		//06 DI6  MX6UL_PAD_ENET2_RX_DATA0__GPIO2_IO08
	IMX_GPIO_NR(2, 9),		//07 DI7  MX6UL_PAD_ENET2_RX_DATA1__GPIO2_IO09
	//IMX_GPIO_NR(2, 3),  	//10 GSM开机
	//IMX_GPIO_NR(2, 3),  	//11 GSM状态
	//IMX_GPIO_NR(2, 3),  	//12 GSM复位
	//IMX_GPIO_NR(2, 3),	//13 
	//IMX_GPIO_NR(2, 3),	//14 
	//IMX_GPIO_NR(1, 6),	//15 POW
	IMX_GPIO_NR(5, 7),		//16 SYS 板载状态灯 MX6UL_PAD_SNVS_TAMPER7__GPIO5_IO07
	IMX_GPIO_NR(1, 5),		//17 DO2 高电平有效 MX6UL_PAD_GPIO1_IO05__GPIO1_IO05
	IMX_GPIO_NR(1, 9),		//18 DO3 高电平有效 MX6UL_PAD_GPIO1_IO05__GPIO1_IO09
	IMX_GPIO_NR(5, 8),		//19 ERR            MX6UL_PAD_SNVS_TAMPER8__GPIO5_IO08
};

/* 用来指定LED所用的GPIO引脚 */
static char gpio_table_name [GPIO_PIN_SIZE][16] =
{
	"DO0",  //0.1 DO输出	    2	GPF3、GPF5	高电平有效
	"DO1",
	"DI0",  //2.9 DI输入	    8	低电平输入有效
	"DI1",
	"DI2",
	"DI3",
	"DI4",
	"DI5",
	"DI6",
	"DI7",
	//"",  //10 GSM开机	    1	GPF7	输出高电平1s，自动GSM开机
	//"",  //11 GSM状态	    1	GPF6	开关量输入，检测GSM的状态
	//"",  //12 GSM复位	    1	GPG1	开关量输出，1复位 0正常
	//"ERR",  //13 
	//"Unkown",  //14 
	//"POW",  //15
	"SYS",  //16 板载状态灯
	"DO2",  //17 DO2输出	    2	GPQ6	高电平有效
	"DO3",  //18 DO2输出	    2	GPQ6	高电平有效
	"ERR",  //19 ERR输出
	//"Unkown",  //
};

/*
 * 用来指定GPIO引脚的功能：输入 输出
 * [1:0]    DO输出
 * [9:2]    DI输入
 * [12:10]  GSM开机，状态，复位
 * [13]     电池充电控制
 * [14]     电池供电控制
 * [15]     电源指示灯控制
 * [16]     系统运行指示灯控制
 */
static unsigned int gpio_cfg_table [GPIO_PIN_SIZE] =
{
    1,  //DO输出	    2	P3.20 P3.21	高电平有效
	1,
	0,  //DI输入	    8	P2.4 P2.5 P2.6 P2.7 P2.12 P2.13 P2.14 P2.15 低电平输入有效
	0,
	0,
	0,
	0,
	0,
	0,
	0,
	//0,  //GSM开机	    1	GPF7	输出高电平1s，自动GSM开机
	//0,  //GSM状态	    1	GPF6	开关量输入，检测GSM的状态
	//0,  //GSM复位	    1	GPG1	开关量输入，1复位 0正常
	//1,  //13 
	//1,  //14 
	//1,  //15 
	1,  //16 板载状态灯
    1,  //17 DO2输出	    1	GPQ6	高电平有效
    1,  //18 DO2输出	    1	GPQ6	高电平有效
	1,  //19 ERR输出
    //1,
};

unsigned int gpio_initval_table[GPIO_PIN_SIZE] =
{
    1, 1,
    0, 0, 0, 0, 0, 0, 0, 0,
    //0, 0, 0,
    //0, 0, 0,
	0,
	1,1,
	0
};

static int gpio_open(struct inode *inode, struct file *filp)
{
	struct user_gpio *dev_info;		/* device information */

	dev_info = container_of(inode->i_cdev, struct user_gpio, gpio_cdev);
	filp->private_data = dev_info;	/* for other methods */

	return 0;
}

static int gpio_release(struct inode *inode, struct file *filp)
{
	return 0;
}

static ssize_t gpio_read(struct file *filp, char __user *buf, size_t count,
                loff_t *f_pos)
{
	int i = 0;
	int ret = 0;
	size_t size = count;
	struct s3c2416_gpio cfg;
//	printk("%s size is %d\n",__func__,size);
	if(size != 4)
		return -EINVAL;

	ret = copy_from_user((void*)&cfg,buf,size);
	//printk("begin: pin is %d  value is %d\n",cfg.pin,cfg.value);

	//if (cfg.pin >= 19)
	//{
	//	return -EINVAL;
	//}

    if (cfg.pin < 11)//gpio单个值
    {
		if (10 == cfg.pin)
			cfg.value = ((gpio_get_value(gpio_table[11])) ? 1 : 0);
		else
			cfg.value = ((gpio_get_value(gpio_table[cfg.pin])) ? 1 : 0);
        ret = copy_to_user(buf,(void*)&cfg,4);
    }
    else if (16 == cfg.pin)//SYS 1bit
    {
        cfg.value = (gpio_get_value(gpio_table[10])) ? 1 : 0;
        ret = copy_to_user(buf,(void*)&cfg,4);
    }
    else if (17 == cfg.pin)//DO 4bit
    {
		cfg.value = 0;
        cfg.value = (cfg.value << 1) | ((gpio_get_value(gpio_table[12])) ? 1 : 0);
        cfg.value = (cfg.value << 1) | ((gpio_get_value(gpio_table[11])) ? 1 : 0);
        cfg.value = (cfg.value << 1) | ((gpio_get_value(gpio_table[1])) ? 1 : 0);
        cfg.value = (cfg.value << 1) | ((gpio_get_value(gpio_table[0])) ? 1 : 0);
        ret = copy_to_user(buf,(void*)&cfg,4);
    }
    else if (18 == cfg.pin)//DI 8bit
    {
        cfg.value = 0;
        for (i = 7; i >= 0; --i)
            cfg.value = (cfg.value << 1) | ((gpio_get_value(gpio_table[2+i])) ? 1 : 0);

        ret = copy_to_user(buf,(void*)&cfg,4);
    }
	else
	{
		return -EINVAL;
	}
	//printk("end: pin is %d  value is %d\n",cfg.pin,cfg.value);

	return ret;
}

ssize_t gpio_write(struct file *filp, const char __user *buf, size_t count,
                loff_t *f_pos)
{
	int ret = 0;
	int value = 0;
	size_t size = count;
	struct s3c2416_gpio cfg;

//	printk("%s size is %d\n",__func__,size);
	if(size != 4)
		return -EINVAL;

	ret = copy_from_user((void*)&cfg,buf,size);
//	printk("pin is %d  value is %d\n",cfg.pin,cfg.value);

	//if (cfg.pin >= 19)
	//{
	//	return -EINVAL;
	//}

    value = (cfg.value ? 1 : 0);
    switch(cfg.pin)
    {
    case 0:
        //break;
    case 1:
        //break;
    //case 10:
        //break;
    //case 12:
        gpio_set_value(gpio_table[cfg.pin],value);
        break;
    //case 13:
        //break;
    //case 14:
        //break;
    //case 15:
        //break;
    case 16:
        //break;
    case 17:
    case 18:
    case 19:
        gpio_set_value(gpio_table[cfg.pin-6],value);
        break;
    default:
        return -EINVAL;
        break;
    }
	//cfg.value = s3c2410_gpio_getpin(gpio_table[cfg.pin]);

	return ret;
}


static loff_t gpio_llseek(struct file *filp, loff_t offset, int origin)
{
	return 0;
}

static long gpio_ioctl(struct file *filp,  unsigned int cmd,  unsigned long arg)
{
	int ret = 0;
	//int i;

	/*user_gpio_arg * parg = (user_gpio_arg*) arg;  
	struct user_gpio *dev_info = (user_gpio *)filp->private_data; 
	
	if (_IOC_TYPE(cmd) != GPIO_IOC_MAGIC) return -ENOTTY;
	if (_IOC_NR(cmd) > GPIO_IOC_MAXNR) return -ENOTTY;

	if (_IOC_DIR(cmd) & _IOC_READ)
		ret = !access_ok(VERIFY_WRITE, (void __user *)arg, _IOC_SIZE(cmd));
	else if (_IOC_DIR(cmd) & _IOC_WRITE)
		ret =  !access_ok(VERIFY_READ, (void __user *)arg, _IOC_SIZE(cmd));

	if (ret) return -EFAULT;

	/// To determine whether the user is GPIO
	for (i = 0; i < dev_info->gpio_num; i++) {
		if (dev_info->gpios[i] == parg->pin) {
			break;
		}
	}

	if (i > dev_info->gpio_num) {
		printk(KERN_WARNING "Unsuported Gpio Num\n");
		return -EFAULT;
	}

	switch (cmd) {

	case IOCTL_GPIO_SETOUTPUT:
			gpio_direction_output(parg->pin, parg->data);
		break;	

	case IOCTL_GPIO_SETINPUT:
			gpio_direction_input(parg->pin);
		break;

	case IOCTL_GPIO_SETVALUE:
			gpio_set_value(parg->pin, parg->data);
		break;

	case IOCTL_GPIO_GETVALUE:
			parg->data = gpio_get_value(parg->pin) ? 1 : 0;
		break;

	default :
		printk(KERN_WARNING "Unsuported Ioctl Cmd: 0x%x\n", cmd);
	}*/
	//printk("cmd = %d  arg = %d\n",cmd, arg);
	switch(cmd)
	{
		case IOCTL_LED_ON:
			// 设置指定引脚的输出电平为0
			switch(arg)
			{
            case 0:
			    //break;
            case 1:
			    //break;
            //case 10:
			    //break;
            //case 12:
				gpio_set_value(gpio_table[arg], 0);
			    break;
            //case 13:
			    //break;
            //case 14:
			    //break;
            //case 15:
			    //break;
            case 16:
			    //break;
            case 17:
            case 18:
            case 19:
				gpio_set_value(gpio_table[arg-6], 0);
			    break;
            default:
                break;
			}
			return 0;

		case IOCTL_LED_OFF:
			// 设置指定引脚的输出电平为0
			switch(arg)
			{
            case 0:
			    //break;
            case 1:
			    //break;
            //case 10:
			    //break;
            //case 12:
				gpio_set_value(gpio_table[arg], 1);
			    break;
            //case 13:
			   //break;
            //case 14:
			    //break;
            //case 15:
			    //break;
            case 16:
			    //break;
            case 17:
            case 18:
            case 19:
				gpio_set_value(gpio_table[arg-6], 1);
			    break;
            default:
                break;
			}
			return 0;

		default:
			return -EINVAL;
	}
	return ret;
}

struct file_operations gpio_fops = {
	.owner =    THIS_MODULE,
	.read =     gpio_read,
	.write =    gpio_write,
	.llseek =   gpio_llseek,
	.open =     gpio_open,
	.unlocked_ioctl =	gpio_ioctl,
	.release =	gpio_release,
};

#if 1
static int user_gpio_probe(struct platform_device *pdev)
{
	int err;
	int i;
	dev_t dev = 0;

	//struct device_node *np = pdev->dev.of_node;
	struct user_gpio *us_gpio;

	us_gpio = kzalloc(sizeof(struct user_gpio), GFP_KERNEL);
	if (!us_gpio) {
		dev_err(&pdev->dev, "Could not allocate memory for platform data\n");
		return -ENOMEM;
	}

	//us_gpio->gpio_num = of_gpio_named_count(np, "pin_gpios");
	us_gpio->gpio_num = GPIO_PIN_SIZE;
	if (us_gpio->gpio_num > 205) {
		err = -EINVAL;
		goto err_free_mem1;
	}
	
	us_gpio->gpios = devm_kzalloc(&pdev->dev, sizeof(unsigned int) * us_gpio->gpio_num, GFP_KERNEL);
	if (!us_gpio->gpios) {
		dev_err(&pdev->dev, "Could not allocate memory for platform data\n");
		err = -ENOMEM;
		goto err_free_mem1;
	}

	printk("GPIO Registered:");
	for (i = 0; i < us_gpio->gpio_num; i++) {
		//us_gpio->gpios[i] = of_get_named_gpio(np, "pin_gpios", i);
		us_gpio->gpios[i] = gpio_table[i];
		printk("%d ", us_gpio->gpios[i]);
	}
	printk("\n");
	/// request for all users GPIO, default gpio set input
	/*for (i = 0; i < us_gpio->gpio_num; i++) {
		err = gpio_request(us_gpio->gpios[i], "user_gpio");
		if (err) {
			dev_err(&pdev->dev, "Failed to request GPIO%d for num %d\n",
				us_gpio->gpios[i], i);
			goto err_free_pin;
		}

		gpio_direction_input(us_gpio->gpios[i]);
	}*/
	for (i = 0; i < GPIO_PIN_SIZE; i++)
	{
		gpio_request(gpio_table[i], gpio_table_name[i]);
		
		if (gpio_cfg_table[i])
			gpio_direction_output(gpio_table[i], gpio_initval_table[i]);
		else
			gpio_direction_input(gpio_table[i]);
	}

	if (gpio_major) {
		dev = MKDEV(gpio_major, gpio_minor);
		err = register_chrdev_region(dev, 1, "user_gpio");
	} else {
		err = alloc_chrdev_region(&dev, gpio_minor, 1, "user_gpio");
		gpio_major = MAJOR(dev);
	}

	if (err < 0) {
		printk(KERN_WARNING "Can't get major %d\n", gpio_major);
		goto err_free_pin;
	}

	cdev_init(&us_gpio->gpio_cdev, &gpio_fops);
	us_gpio->gpio_cdev.owner = THIS_MODULE;
	us_gpio->gpio_cdev.ops = &gpio_fops;
	err = cdev_add(&us_gpio->gpio_cdev, dev, 1);

	/// derves node file
	gpio_class = class_create(THIS_MODULE, "user_gpio");
	if (IS_ERR(gpio_class)) {
		printk(KERN_WARNING "class_create faild\n");
		cdev_del(&us_gpio->gpio_cdev);
		unregister_chrdev_region(dev, 0);
		goto err_free_pin;
	}

	class_dev = device_create(gpio_class, NULL, dev, 0, "gpio");

	platform_set_drvdata(pdev, us_gpio);

	return 0;
err_free_pin:
	/*for (i--; i >= 0; i--) {
		gpio_free(us_gpio->gpios[i]);
	}*/
	for (i = 0; i < GPIO_PIN_SIZE; i++)
	{
		gpio_free(gpio_table[i]);
	}

	kfree(us_gpio->gpios);
err_free_mem1:
	kfree(us_gpio);

	return err;
}


static void user_gpio_remove(struct platform_device *pdev)
{
	/// gpio free
	int i;
	struct user_gpio *us_gpio = platform_get_drvdata(pdev);
	/*for (i = 0; i < us_gpio->gpio_num; i++) {
		gpio_free(us_gpio->gpios[i]);
	}*/
	for (i = 0; i < GPIO_PIN_SIZE; i++)
	{
		gpio_free(gpio_table[i]);
	}

	if (!us_gpio->gpios)
		kfree(us_gpio->gpios);
	if (!us_gpio)
		kfree(us_gpio);
}


static struct of_device_id user_gpio_of_match[] = {
	{ .compatible = "user_gpio", },
	{ },
};

static struct platform_driver user_gpio_device_driver = {
	.probe		= user_gpio_probe,
	.remove		= user_gpio_remove,
	.driver		= {
		.name	= "gpio",
		.owner	= THIS_MODULE,
		.of_match_table = of_match_ptr(user_gpio_of_match),
	}
};

static int __init user_gpio_init(void)
{
	return platform_driver_register(&user_gpio_device_driver);
}

static void __exit user_gpio_exit(void)
{
	dev_t dev = 0;

	dev = MKDEV(gpio_major, gpio_minor);

	device_del(class_dev);
	class_destroy(gpio_class);
	unregister_chrdev_region(dev, 0); 
	///cdev_del(&gpio_cdev);
}
#else
#include <linux/cdev.h>
#include <linux/miscdevice.h>
static struct miscdevice misc = {
	.minor = MISC_DYNAMIC_MINOR,
	.name = "gpio",
	.fops = &gpio_fops,
};

static int __init user_gpio_init(void)
{
	int ret;
	int i;
	//dev_t devno = MKDEV(tp_MAJOR,0);
	//struct class *gpio_class;
	for (i = 0; i < GPIO_PIN_SIZE; i++)
	{
		gpio_request(gpio_table[i], gpio_table_name[i]);
		
		if (gpio_cfg_table[i])
			gpio_direction_output(gpio_table[i], gpio_initval_table[i]);
		else
			gpio_direction_input(gpio_table[i]);
	}
	ret = misc_register(&misc);

	return ret;
}

static void __exit user_gpio_exit(void)
{
	dev_t dev = 0;
	int i=0;
	for (i = 0; i < GPIO_PIN_SIZE; i++)
	{
		gpio_free(gpio_table[i]);
	}
	dev = MKDEV(gpio_major, gpio_minor);

	device_del(class_dev);
	class_destroy(gpio_class);
	unregister_chrdev_region(dev, 0); 
	///cdev_del(&gpio_cdev);
}
#endif

module_init(user_gpio_init);

module_exit(user_gpio_exit);


MODULE_DESCRIPTION("Driver for User GPIO");
MODULE_AUTHOR("peixiuhui@163.com");
MODULE_LICENSE("GPL");

